import sys, pygame, math, random, copy
from pygame.locals import *

HEADERHEIGHT = 40
HEADERTEXTHEIGHT = 30
BORDERWIDTH = 2
WORLDWIDTH = 600
WIDTH, HEIGHT = BORDERWIDTH*2+WORLDWIDTH, BORDERWIDTH*2+HEADERHEIGHT+WORLDWIDTH

GRIDDIM = 20
FPS = 25

JUMPSPEED = 3.0
RUNSPEED = 1.5
GRAVITY = 3.0
SPEEDUP = 3

SPARSENESS = 1 #0.99

def sgn(x):
	if x > 0:
		return 1
	elif x < 0:
		return -1
	else:
		return 0

class Transform:
	wtscache = None

	@staticmethod
	def worldToScreen((x,y)):
		"Transforms the world coordinate to screen coordinates"
		# (0,0) => (BORDERWIDTH,HEIGHT-BORDERWIDTH)
		# (GRIDDIM,GRIDDIM) => (BORDERWIDTH+WORLDWIDTH,HEIGHT-BORDERWIDTH-WORLDWIDTH)
		# (x,y) between (0,0) and (1,1) => (BORDERWIDTH+x/GRIDDIM*WORLDWIDTH,HEIGHT-BORDERWIDTH-y/GRIDDIM*WORLDWIDTH)
	
		# cache some of the calculations
		if not Transform.wtscache:
			Transform.wtscache = ((1.0/GRIDDIM)*WORLDWIDTH,HEIGHT-BORDERWIDTH)
		
		return (BORDERWIDTH+int(x*Transform.wtscache[0]),Transform.wtscache[1]-int(y*Transform.wtscache[0]))

	@staticmethod
	def worldToScreenSize((w,h)):
		if not Transform.wtscache:
			Transform.wtscache = ((1.0/GRIDDIM)*WORLDWIDTH,HEIGHT-BORDERWIDTH)

		return (w*Transform.wtscache[0],h*Transform.wtscache[0])
	
BLOCKWIDTH = Transform.worldToScreen((1,0))[0] - Transform.worldToScreen((0,0))[0]


class Block:
	# type
	Empty, Normal, Growing = range(3)
	# sides
	Left, Right, Top, Bottom = range(4)
	
	def __init__(self,i,j,t=Empty):
		self.i = i
		self.j = j
		self.typ = t
		self.isdudeon = False
		self.neighbours = []
		if i>0:
			self.neighbours.append((i-1,j))
		if i<(GRIDDIM-1):
			self.neighbours.append((i+1,j))
		if j>0:
			self.neighbours.append((i,j-1))
		if j<(GRIDDIM-1):
			self.neighbours.append((i,j+1))
		return

	def settype(self,t):
		self.typ = t
	def gettype(self):
		return self.typ
	def dudeon(self):
		return self.isdudeon
	def setdudeon(self,don):
		self.isdudeon = don
		return don

	def getneighbours(self):
		return self.neighbours
		

class World:
	rng = random.Random(123)
	
	theblocks = [[Block(i,j) for i in range(GRIDDIM)] for j in range(GRIDDIM)]
	activeBlocks = []
	growingBlocks = []
	
	# activeBlocksRects = [] # cached rectangles of active blocks
	currentT = 0
	GROWTHRATE = 500 # in ms
	BIRTHRATE = 1200

	@staticmethod
	def block(i,j):
		return World.theblocks[j][i]

	@staticmethod
	def blocks():
		for i in range(GRIDDIM):
			for j in range(GRIDDIM):
				yield World.block(i,j)

	@staticmethod
	def activeblocks():
		for block in World.activeBlocks:
			yield block
	
	@staticmethod
	def newblock(i,j,t):
		bl = World.block(i,j)
		bl.settype(t)
		World.activeBlocks.append(bl)

	@staticmethod
	def clear():
		for i in range(GRIDDIM):
			for j in range(GRIDDIM):
				World.block(i,j).settype(Block.Empty)

	@staticmethod
	def evolve(dt):
		"transforms the world according to stochastic growth rules"
		World.currentT += dt
		
		# update growing blocks
		toremove = []
		for blt in World.growingBlocks:			
			blt[1] += dt			
			if blt[1]>World.BIRTHRATE:
				blt[0].settype(Block.Normal)
				toremove.append(blt)
				World.activeBlocks.append(blt[0])
				
		for blt in toremove:
			World.growingBlocks.remove(blt)
		
		if World.currentT > World.GROWTHRATE:

			# identify squares we can grow into
			accessible_squares = []
			for block in World.blocks():
				nb = [World.block(x[0],x[1]).gettype() for x in block.getneighbours()]
				nbe = nb.count(Block.Empty)
				if block.gettype()==Block.Empty and nbe==len(nb)-1:
					accessible_squares.append(block)

			if len(accessible_squares)==0:
				print "No more empty spots!"
				return

			# select one at random
			nextblock = accessible_squares[World.rng.randint(0,len(accessible_squares)-1)]
			World.growingBlocks.append([nextblock,World.currentT-World.GROWTHRATE])
			nextblock.settype(Block.Growing)

			# start growing a new block
##			i,j = random.randint(0,GRIDDIM-1), random.randint(0,GRIDDIM-1)
##			if World.block(i,j).gettype()==Block.Empty:
##				World.growingBlocks.append([(i,j),World.currentT-World.GROWTHRATE])
##				World.block(i,j).settype(Block.Growing)

			World.currentT -= World.GROWTHRATE


gScreen, gClock = [None]*2
gTotalTime = 0
gMenuFont = None
gValidKeys = [K_ESCAPE, K_DOWN, K_UP, K_LEFT, K_RIGHT, K_SPACE, K_RETURN]
gKeys = {}
for key in gValidKeys:
	gKeys[key] = False

# Mode enum
class Mode:
	menu, start, play, pause, dead, highscore, quitt = range(7)
	mode = menu
	
# Action enum
class Action:
	left, right, jump, up, down, select = range(6)

# mode logic
class Menu:
	play, info, quitt = range(3)
	labels = ["!","?","x"]
	
gMenuItemSelected = Menu.play
gMenuItemSurfaces = []
gMenuItemPositions = [[0,0],[0,0],[0,0]]

class MenuMode:
	idle, down, up, selected, info = range(5)

gMenuMode = MenuMode.idle
gMenuScrollSpeed = 100 #ms per item
gMenuScrollCounter = 0
gKeyLock = False #lock escape key from quitting directly after escaping to me

def menu():
	global gMenuItemsSurfaces, gMenuItemSelected, gMenuItemPositions, gMenuScrollCounter, gMenuMode, gKeyLock
	
	# update mode
	if gMenuMode == MenuMode.idle:	
		if gKeys[K_ESCAPE] and not gKeyLock:
			Mode.mode = Mode.quitt
			return
		elif not gKeys[K_ESCAPE] and gKeyLock:
			gKeyLock = False
		elif gKeys[K_DOWN]:
			gMenuMode = MenuMode.down
			gMenuScrollCounter = gMenuScrollSpeed+1
		elif gKeys[K_UP]:
			gMenuMode = MenuMode.up
			gMenuScrollCounter = gMenuScrollSpeed+1
		elif gKeys[K_SPACE]:
			gMenuMode = MenuMode.selected
	elif gMenuMode == MenuMode.down:
		if not gKeys[K_DOWN]:
			gMenuMode = MenuMode.idle
		gMenuScrollCounter += gClock.get_time()
		if gMenuScrollCounter>gMenuScrollSpeed:
			gMenuScrollCounter = 0
			gMenuItemSelected = (gMenuItemSelected + 1) % len(Menu.labels)
	elif gMenuMode == MenuMode.up:
		if not gKeys[K_UP]:
			gMenuMode = MenuMode.idle
		gMenuScrollCounter += gClock.get_time()
		if gMenuScrollCounter>gMenuScrollSpeed:
			gMenuScrollCounter = 0
			gMenuItemSelected = (gMenuItemSelected - 1) % len(Menu.labels)
	elif gMenuMode == MenuMode.selected:
		if gMenuItemSelected == Menu.quitt:
			Mode.mode = Mode.quitt
			return
		elif gMenuItemSelected == Menu.play:
			Mode.mode = Mode.play
			return
		elif gMenuItemSelected == Menu.info:
			gMenuMode = MenuMode.info			
	elif gMenuMode == MenuMode.info:
		if gKeys[K_ESCAPE]:
			gMenuMode = MenuMode.idle
			gKeyLock = True
	
	# draw menu
	if not gMenuMode == MenuMode.info:
		ht = gMenuFont.get_height()
		totalmenuheight = len(gMenuItemSurfaces)*ht
		pos = [WIDTH/2,(HEIGHT-totalmenuheight)/2]
		for i in range(len(gMenuItemSurfaces)):
			item = gMenuItemSurfaces[i]
			w = item.get_width()
			gMenuItemPositions[i] = [pos[0]-w/2,pos[1]+ht*.04*math.sin(gTotalTime/1000.0)]
			gScreen.blit(item,(gMenuItemPositions[i][0],gMenuItemPositions[i][1]))
			pos[1] += ht

		# highlight selected item
		srf = gMenuItemSurfaces[gMenuItemSelected]
		border = 10	
		pygame.draw.rect(
			gScreen,
			(255,255,255),
			pygame.Rect(
				gMenuItemPositions[gMenuItemSelected][0]-border,
				gMenuItemPositions[gMenuItemSelected][1]-border,
				srf.get_width()+2*border,
				srf.get_height()+2*border),
			1)
	else:
		pygame.draw.rect(
			gScreen,
			(255,0,0),
			pygame.Rect(10,10,WIDTH-20,HEIGHT-20))

gNumberSurfaces = []
def renderDigit(i,pos):
	"assumes i is in [0,9]"
	return gScreen.blit(gNumberSurfaces[i],pos)

def renderInteger(num,pos,paddedwidth=0):
	dlist = [int(x) for x in str(num)] #list of digits
	p = list(pos)
	nzeros = paddedwidth - len(dlist)
	if nzeros > 0:
		while nzeros>0:
			rct = renderDigit(0,p)
			p[0] += rct.width
			nzeros = nzeros-1
		
	for i in dlist:
		rct = renderDigit(i,p)
		p[0] += rct.width

def renderIntegerBounds(num,paddedwidth=0):
	"returns the screen width of the number"
	dlist = [int(x) for x in str(num)] #list of digits
	nzeros = paddedwidth - len(dlist)
	width = 0
	height = 0
	if nzeros > 0:
		width += gNumberSurfaces[0].get_width()*nzeros
		height = max(height,gNumberSurfaces[0].get_height())
	for i in dlist:
		width += gNumberSurfaces[i].get_width()
		height = max(height,gNumberSurfaces[0].get_height())
		
	return (width,height)

# Primary game logic

# The game starts with the player standing on the first block
# Each DT seconds a new block is added to the arena
# Before it is added a rectangle flashes so the player knows where it is appearing
# If the block appears in the same location as the player, the player dies
# The player can jump and move left or right
#   A simple physics and collision detection governs the dynamics of the world
# Blocks disappear after a short time, and no new blocks can appear there
# The game ends if the player dies or reaches the top right corner of the world

# Additional features:
# - blocks have special properties, denoted by their colour
# - white blocks are normal
# - e.g., green blocks enable extra high jumps, blue blocks have no friction
# -       orange blocks are unstable
# - the player can double jump, wall jump, long jump, etc. (feature creep!?)

# COORDINATE SYSTEM
# bottom-left corner of world = (0,0)
# top-right corner = (GRIDDIM,GRIDDIM)

class PlayMode:
	init, starting, playing, won, lost = range(5)
	mode = init

class Dude:
	wsz = .4,.66
	ssz = Transform.worldToScreenSize(wsz)
	pos = [GRIDDIM/2.0+0.5-wsz[0]/2,GRIDDIM/2.0 + 1] #bottom left point of dude
	vel = [0,0]
	airborn = False
	onblock = None
	collided = False #has he just collided mid-air? affects jumping ability

	@staticmethod
	def right():
		return Dude.pos[0] + Dude.wsz[0]
	@staticmethod
	def left():
		return Dude.pos[0]
	@staticmethod
	def top():
		return Dude.pos[1] + Dude.wsz[1]
	@staticmethod
	def bottom():
		return Dude.pos[1]
	
	@staticmethod
	def draw():
		sc = Transform.worldToScreen(Dude.pos)
		pygame.draw.rect(gScreen,(50,50,100),pygame.Rect(sc[0],sc[1]-Dude.ssz[1],Dude.ssz[0],Dude.ssz[1]))

	@staticmethod
	def jump():
		if not Dude.airborn:
			Dude.vel[1] = JUMPSPEED

	@staticmethod
	def run(hv):
		if not Dude.airborn:
			Dude.vel[0] = hv
		elif not Dude.collided:			
			if Dude.vel[0]>0 and hv < 0:
				Dude.vel[0] *= .8
				Dude.vel[0] = max(Dude.vel[0],0.1)
			elif Dude.vel[0]<0 and hv > 0:
				Dude.vel[0] *= .8
				Dude.vel[0] = min(Dude.vel[0],-0.1)
			else:
				Dude.vel[0] = hv
		
	@staticmethod
	def update():
		dt = gClock.get_time()*SPEEDUP/1000.0
		if Dude.airborn:
			Dude.vel[1] -= dt*GRAVITY
			Dude.pos[0] += dt*Dude.vel[0]
			Dude.pos[1] += dt*Dude.vel[1]

			# sides of bounding rectangle
			left, right = Dude.pos[0], Dude.pos[0] + Dude.wsz[0]
			top, bottom = Dude.pos[1]+Dude.wsz[1], Dude.pos[1]
			
			# first with the boundaries of the world	
			if left < 0:
				Dude.pos[0] = 0
				Dude.vel[0] = 0
			elif right > GRIDDIM:
				Dude.pos[0] = GRIDDIM - Dude.wsz[0]
				Dude.vel[0] = 0
				
			if top > GRIDDIM:
				# hit ceiling, thus won game
				Dude.pos[1] = GRIDDIM - Dude.wsz[1]
				Dude.vel[1] = 0
				PlayMode.mode = PlayMode.won
			elif bottom < 0:
				# hit floor, thus dead
				Dude.pos[1] = 0
				Dude.vel[1] = 0
				Dude.airborn = False
				PlayMode.mode = PlayMode.lost

			# detect collision with active blocks
			# XXX: We don't need to check all active blocks, as we have the coordinates of the guy!
			# only need to check the block of 4 surrounding the guy!
			for block in World.activeblocks():
				bleft, bright = block.i, block.i+1
				btop, bbottom = block.j+1, block.j

				if right < bleft: continue
				if left > bright: continue
				if top < bbottom: continue
				if bottom > btop: continue

				# hence a collision has occurred
				side = None

				# calculate dx and dy
				# using the leading edges (based on vel)

				dx, dy = 0, 0
				if Dude.vel[0] < 0:
					dx = bright - left
				elif Dude.vel[0] > 0:
					dx = right - bleft

				if Dude.vel[1] < 0:
					dy = btop - bottom
				elif Dude.vel[1] > 0:
					dy = top - bbottom

				#print "i", dx, dy, Dude.vel

				vellen = math.sqrt(Dude.vel[0]*Dude.vel[0] + Dude.vel[1]*Dude.vel[1])
				velnorm = -Dude.vel[0]/vellen, -Dude.vel[1]/vellen

				t = 0				
				if Dude.vel[0]==0:
					# Note Dude.vel != [0,0], hence
					# t  = ty
					t = dy / math.fabs(velnorm[1])
					if Dude.vel[1]>0:
						side = Block.Bottom
					else:
						side = Block.Top
				elif Dude.vel[1]==0:
					t = dx / math.fabs(velnorm[0])
					if Dude.vel[0] < 0:
						side = Block.Right
					else:
						side = Block.Left
				else:
					tx = dx / math.fabs(velnorm[0])
					ty = dy / math.fabs(velnorm[1])
					#print "t",tx,ty
					if tx<ty:						
						t = tx
						if Dude.vel[0]<0:
							side = Block.Right
						else:
							side = Block.Left
					else:
						t = ty
						if Dude.vel[1]>0:
							side = Block.Bottom
						else:
							side = Block.Top
							
				epsilon = 0.001 #extra amount to shift dude by	
				disp = (velnorm[0]+sgn(velnorm[0])*epsilon)*t, (velnorm[1]+sgn(velnorm[1])*epsilon)*t
				#print "d", disp
				Dude.pos[0] += disp[0]
				Dude.pos[1] += disp[1]

				# print disp
				if side==Block.Top:
					block.setdudeon(True)
					Dude.airborn = False
					Dude.onblock = block
					Dude.vel = [Dude.vel[0],0]
				elif side==Block.Bottom:
					Dude.vel = [Dude.vel[0],0]
				elif side==Block.Left or side==Block.Right:
					Dude.vel = [0,Dude.vel[1]]

				if not side==Block.Top:
					Dude.collided = True
				else:
					Dude.collided = False
				
				break
				
		
		else:
			if Dude.vel[1] > 0:
				Dude.airborn = True
				if Dude.onblock:
					Dude.onblock.setdudeon(False)
					Dude.onblock = None
			else:
				Dude.pos[0] += dt*Dude.vel[0]
				eps = 0.0001

				# bound to play area
				if Dude.right() > GRIDDIM:
					Dude.pos[0] = GRIDDIM - Dude.wsz[0] - eps
					return
				elif Dude.left() < 0:
					Dude.pos[0] = eps
					return

				# check collision with block
				# only need to check current row
				row = 0
				if Dude.onblock:
					row = Dude.onblock.j + 1
				for i in range(0,GRIDDIM):                              
					bl = World.block(i,row)
					if not bl.gettype()==Block.Empty and not bl.gettype()==Block.Growing:
						if Dude.left() < bl.i+1 and Dude.right() > bl.i:
							#       then a collision has occurred
							# push back dude in opposite velocity direction
	
							if Dude.vel[0] > 0:
								Dude.pos[0] = i - Dude.wsz[0] - eps
							else:
								Dude.pos[0] = i + 1 + eps
							Dude.vel[0] = 0
							return

				# check falloff block

				if Dude.onblock:
					ob = Dude.onblock
					if Dude.pos[0]+Dude.wsz[0] < ob.i:
						bl = World.block(ob.i-1,ob.j)
						if bl.gettype()==Block.Empty:
							Dude.airborn = True
							ob.setdudeon(False)
							Dude.onblock = None
						else:
							ob.setdudeon(False)
							Dude.onblock = bl
							bl.setdudeon(True)
					elif Dude.pos[0] > ob.i + 1:
						bl = World.block(ob.i+1,ob.j)
						if bl.gettype()==Block.Empty:
							Dude.airborn = True
							ob.setdudeon(False)
							Dude.onblock = None
						else:
							ob.setdudeon(False)
							Dude.onblock = bl
							bl.setdudeon(True)
						
				#Dude.pos[1] += dt*Dude.vel[1]
				#
				Dude.vel[0] *= 0.5


	
def play():
	global gWorld, gActiveBlocks, gActiveBlocksRects
	
	# update mode

	if PlayMode.mode==PlayMode.init:
		# initialise
		World.clear()
		for i in range(GRIDDIM):
			for j in range(GRIDDIM):
				if random.random()>SPARSENESS or i==GRIDDIM/2 and j==3:
					World.newblock(i,j,Block.Normal)
		Dude.airborn = True
		PlayMode.mode = PlayMode.starting
		
	elif PlayMode.mode==PlayMode.starting:
		# starting animation

		PlayMode.mode = PlayMode.playing
	elif PlayMode.mode==PlayMode.playing:	
		# handle input
		if gKeys[K_ESCAPE]:
			Mode.mode = Mode.quitt
			return
		
		if gKeys[K_LEFT]:
			Dude.run(-RUNSPEED)
		elif gKeys[K_RIGHT]:
			Dude.run(RUNSPEED)

		if gKeys[K_UP]:
			Dude.jump()

		# step physics
		Dude.update()

		# check status of game, update() may have changed it
		if not PlayMode.mode==PlayMode.playing:
			return	

		# step world evolution
		World.evolve(gClock.get_time())

		# check for death


		# draw game

		# draw header
		pygame.draw.rect(gScreen,(50,50,50),pygame.Rect(0,0,WIDTH,HEADERHEIGHT))
		clkw, clkh = renderIntegerBounds(0,3)
		renderInteger(int(gTotalTime/1000),(0,0),3)
		
		# draw world
		blockcolor = (255,255,255)
		for block in World.blocks():
			si, sj = Transform.worldToScreen((block.i,block.j+1))
			rect = pygame.Rect(si,sj,BLOCKWIDTH,BLOCKWIDTH)
			if block.gettype()==Block.Normal:
				pygame.draw.rect(gScreen,blockcolor,rect)
			elif block.gettype()==Block.Growing:
				pygame.draw.rect(gScreen,blockcolor,rect,1)
				
		# draw dude
		Dude.draw()
	elif PlayMode.mode == PlayMode.won:
		# handle input
		if gKeys[K_ESCAPE]:
			Mode.mode = Mode.quitt
			return
		
		# do win sequence
		
		# draw header
		pygame.draw.rect(gScreen,(0,0,150),pygame.Rect(0,0,WIDTH,HEADERHEIGHT))
		clkw, clkh = renderIntegerBounds(0,3)
		renderInteger(int(gTotalTime/1000),(0,0),3)
		
		# draw world
		blockcolor = (255,255,255)
		for block in World.blocks():
			si, sj = Transform.worldToScreen((block.i,block.j+1))
			rect = pygame.Rect(si,sj,BLOCKWIDTH,BLOCKWIDTH)
			if block.gettype()==Block.Normal:
				pygame.draw.rect(gScreen,blockcolor,rect)
			elif block.gettype()==Block.Growing:
				pygame.draw.rect(gScreen,blockcolor,rect,1)
				
		# draw dude
		Dude.draw()
		
	elif PlayMode.mode == PlayMode.lost:
		# handle input
		if gKeys[K_ESCAPE]:
			Mode.mode = Mode.quitt
			return
		
		# draw header
		pygame.draw.rect(gScreen,(150,0,0),pygame.Rect(0,0,WIDTH,HEADERHEIGHT))
		clkw, clkh = renderIntegerBounds(0,3)
		renderInteger(int(gTotalTime/1000),(0,0),3)
		
		# draw world
		blockcolor = (255,255,255)
		for block in World.blocks():
			si, sj = Transform.worldToScreen((block.i,block.j+1))
			rect = pygame.Rect(si,sj,BLOCKWIDTH,BLOCKWIDTH)
			if block.gettype()==Block.Normal:
				pygame.draw.rect(gScreen,blockcolor,rect)
			elif block.gettype()==Block.Growing:
				pygame.draw.rect(gScreen,blockcolor,rect,1)
				
		# draw dude
		Dude.draw()

def dead():
	None
	# enter dead logic	

#key handler
def keydown(key):
	gKeys[key] = True

def keyup(key):
	gKeys[key] = False

# Main
def main():

	setup()
	
	finished = False	
	# main loop 
	while not finished:
		# handle events		
		for event in pygame.event.get():
			# quit game
			if event.type == QUIT:
				finished = True
				break
			elif event.type == KEYDOWN:
				keydown(event.key)
			elif event.type == KEYUP:
				keyup(event.key)
				
		gScreen.fill((0,0,0))
		# take a step in current mode
		{Mode.menu: menu,
		 Mode.play: play,
		 Mode.dead: dead}[Mode.mode]()

		# flip the screen and wait
		pygame.display.flip()
		tick(FPS)
			
		if Mode.mode==Mode.quitt:
			finished = True

	# handle quitting
	pygame.quit()

def tick(fps):
	global gTotalTime
	gClock.tick(1000.0/fps)
	gTotalTime += gClock.get_time()

def setup():
	"Setup cairo and pygame"
	global gScreen, gClock, gTotalTime, gMenuFont, gMenuItemSurfaces, gMenuItemSelected, gNumberSurfaces

	pygame.init()
	gScreen = pygame.display.set_mode((WIDTH, HEIGHT))
	
	gTotalTime = 0
	gClock = pygame.time.Clock()

	# setup menu and text
	gMenuFont = pygame.font.Font(pygame.font.get_default_font(),HEIGHT/30)

	for item in Menu.labels:
		gMenuItemSurfaces.append(gMenuFont.render(item,True,(255,255,255)))
	gMenuItemSelected = Menu.play

	numberFont = pygame.font.Font(pygame.font.get_default_font(),HEADERTEXTHEIGHT)
	for i in [str(x) for x in range(10)]:
		gNumberSurfaces.append(numberFont.render(i,True,(255,255,255)))
	

if __name__ == '__main__': main()
